/* Mail to web gateway. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>

#include "shoheiutil.h"

/* enable user homepages */
#define HOMEPAGES

/* Read recipients from message, deliver to queue */
#define SENDMAIL	"/usr/lib/sendmail -t -O DeliveryMode=q"
#define INEWS		"/usr/bin/inews"

/* The whereabouts of this program */
#if 0
#define SCRIPT		"/cgi-bin/shohei"
#else
#define SCRIPT		"/"
#endif

/* What to put at the beginning of Subject: when replying */
#define RE		"Re: "

/* What to use for quoted material when replying */
#define QUOTE		"> "

#define INBOX		"INBOX"
#define TRASH		"deleted-messages"

/* Default message id */
#define MSGID		"MSGID"

static int useicons = 0;
static int leaveonserver = 0;
static int adminprivs = 0;
static int superprivs = 0;
static char *language;

static int header_done = 0, footer_done = 0;

#define HEADER 1
#define BODY 2

typedef struct {
	char *from;
	char *subject;
	char *to;
	char *cc;
	char *date;
	char *replyto;
	char *msgid;
	char *ctype;
	char *ctenc;
	size_t offset;
} header;

struct {
	char *attr;
	char *value;
} query[1000];
int nquery;

int msgno[10000];

header headers[1000];
int nmsg;

char *query_string;
char *username, *password, *domain, *hostname, *folder, *ticket, *msg;
char *boundary, *fname;
char dn[1000];

/* attachment */
void *filedata;
int filesize;

int debugcgi = 0;

static void html_error(char *);
static void html_subst(char *);
static void html_template(char *);
static char *require_query(char *);

/*
Generate a pseudorandom ticket name. Tickets are stored in
the directory TICKETDIR.
*/
static void ticketname(void)
{
	char fn[1000], b[1000];
	time_t t0;
	unsigned int t = time(&t0), r;
	FILE *fp;

	srand((unsigned int)getpid());
	for (;;) {
		r = 1000000.0*rand()/(RAND_MAX+1.0);
		csnprintf(b, sizeof b, "%d.%x", t, r);
		csnprintf(fn, sizeof fn, "%s/%s", TICKETDIR, b);
		fp = zopen(fn, "r");
		if (fp == NULL) break;
		fclose(fp);
	}
	ticket = dupstring(b);
}

/*
The "ticket" is used as key to a file which contains the username, password,
hostname and domain. This avoids having the password appear in URL:s
and generated documents. The file is deleted when the user logs out,
which prevents mail theft by guessing tickets.
More values: folder, language, useicons.
*/
static void readticket(void)
{
	FILE *fp;
	char b[1000];
	ticket = require_query("ticket");
	csnprintf(b, sizeof b, "%s/%s", TICKETDIR, ticket);
	fp = zopen(b, "r");
	if (fp == NULL)
		html_error(_("Can't read ticket"));
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read user name from ticket"));
	chomp(b);
	username = dupstring(b);
	if (!fgets(dn, sizeof dn, fp))
		html_error(_("Can't read distinguished name from ticket"));
	chomp(dn);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read password from ticket"));
	chomp(b);
	password = dupstring(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read hostname from ticket"));
	chomp(b);
	hostname = dupstring(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read domain from ticket"));
	chomp(b);
	domain = dupstring(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read folder from ticket"));
	chomp(b);
	folder = dupstring(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read language from ticket"));
	chomp(b);
	language = dupstring(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read useicons from ticket"));
	chomp(b);
	useicons = atoi(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read leaveonserver from ticket"));
	chomp(b);
	leaveonserver = atoi(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read admin from ticket"));
	chomp(b);
	adminprivs = atoi(b);
	if (!fgets(b, sizeof b, fp))
		html_error(_("Can't read super from ticket"));
	chomp(b);
	superprivs = atoi(b);
	fclose(fp);
}

static void chkticket()
{
	if (!ticket) html_error(_("No ticket"));
}

/*
Replace all characters in blacklist. Used to sanitize filenames.
*/
static void blacklist(char *p, char *bl, int r)
{
	while (*p) {
		if (strchr(bl, *p)) *p = r;
		p++;
	}
}

static void writeticket(void)
{
	FILE *fp;
	char b[1000];
	ticketname();
	csnprintf(b, sizeof b, "%s/%s", TICKETDIR, ticket);
	blacklist(username, "/.@", '_');
	blacklist(hostname, "/@", '-');
	blacklist(domain, "/@", '-');
	blacklist(folder, "/@", '_');
	blacklist(language, "/@.", '_');
	if (debugcgi) {
		printf("%s\n%s\n%s\n%s\n%s\n%s\n%s\n%d\n%d\n%d\n",
			username, dn, password, hostname, domain,
			folder, language, useicons, leaveonserver, adminprivs);
	}
	fp = fopen(b, "w");
	if (fp == NULL) html_error(_("Can't write ticket"));
	fprintf(fp, "%s\n%s\n%s\n", username, dn, password);
	fprintf(fp, "%s\n%s\n", hostname, domain);
	fprintf(fp, "%s\n", folder);
	fprintf(fp, "%s\n", language);
	fprintf(fp, "%d\n", useicons);
	fprintf(fp, "%d\n", leaveonserver);
	fprintf(fp, "%d\n", adminprivs);
	fprintf(fp, "%d\n", superprivs);
	fclose(fp);
}

char *make_html(char *p)
{
	char b[2000];
	char *q = b;
	int c;
	if (!p) return dupstring("");
	*q = '\0';
	while ((c = *p++) && ((q-b) < (sizeof b)-10)) {
		if (c == '<') {
			strcat(q, "&lt;");
		} else if (c == '>') {
			strcat(q, "&gt;");
		} else if (c == '"') {
			strcat(q, "&quot;");
		} else if (c == '&') {
			strcat(q, "&amp;");
		} else {
			*q++ = c;
			*q = '\0';
		}
		q += strlen(q);
	}
	return dupstring(b);
}

void printhtml(FILE *fp, char *p)
{
	char *q = make_html(p), *r;
	int i;
	i = 0;
	r = q;
	while (*q) {
		if (i > 70 && *q == ' ') {
			putc('\n', fp);
			i = 0;
		} else {
			putc(*q, fp);
			i++;
		}
		q++;
	}
	free(r);
}

/* Lose the mime crap in headers */
static void printheader(FILE *fp, char *p)
{
	int c, state = 0;
	char *q = p, *r, b[3];

	if (p == NULL) return;

	r = make_html(p);
	p = q = r;
	while ((c = *p++)) {
		if (c == '<') {
			printf("&lt;");
			continue;
		}
		switch (state) {
		case 0:
			if (c == '=') {
				state = 1;
				q = p;
			} else if (c == '<') {
				printf("&lt;");
			} else if (c == '>') {
				printf("&gt;");
			} else {
				putc(c, fp);
			}
			break;
		case 1:
			if (c == '?') {
				state = 2;
			} else {
				state = 0;
				putc('=', fp);
				p = q;
			}
			break;
		case 2:
			if (c == '?') state = 3;
			break;
		case 3:
			if (c == '?') state = 4;
			break;
		case 4:
			if (c == '=') {
				state = 5;
			} else if (c == '?') {
				state = 7;
				q = p;
			} else {
				putc(c, fp);
			}
			break;
		case 5:
			state = 6;
			b[0] = c;
			break;
		case 6:
			state = 4;
			b[1] = c;
			b[2] = '\0';
			putc((int)strtol(b, NULL, 16), fp);
			break;
		case 7:
			if (c == '=') {
				state = 0;
			} else {
				state = 4;
				putc('?', fp);
				p = q;
			}
			break;
		}
	}
	free(r);
}

static void icon(char *img, char *alt)
{
	if (useicons) {
		printf("<img src=\"%s/%s\" border=0 alt=\"%s\">",
			IMGDIR, img, alt);
	} else {
		printf("[%s]", alt);
	}
}


static void html_error(char *p)
{
	if (!header_done) html_subst("HEADER");
	if (p) printf("<hr size=1 noshade>%s<hr size=1 noshade>\n", p);
	printf("%s\n"
		"%s<br>\n"
		"<a href=mailto:ulric@edu.stockholm.se>\n"
		"Ulric Eriksson (ulric@edu.stockholm.se)\n"
		"</a><hr size=1 noshade>\n",
		_("There was a problem processing the request."),
		_("Report any errors or omissions to:"));
	printf("<a href=\"%s\">%s</a>\n", SCRIPT, _("Log in"));
	html_subst("FOOTER");
	exit(0);
}

static long message_size(int n)
{
	return headers[n+1].offset-headers[n].offset;
}

/* Display header for message n. Must run parsemail first. */
static void html_top(int n)
{
	int i = n-1;
	if (headers[i].from == NULL)
		headers[i].from = dupstring(_("No sender"));
	if (headers[i].subject == NULL)
		headers[i].subject = dupstring(_("No subject"));
	printf("<tr bgcolor=\"%s\">\n", (i&1)?"#eeeeee":"#dddddd");
	printf("<td><input type=checkbox name=msg%d></td>\n", i+1);
	printf("<td nowrap>");
	if (headers[i].date) printhtml(stdout, short_date(headers[i].date));
	else printf("&nbsp;");
	printf("</td>\n");
	printf("<td>");
	printhtml(stdout, headers[i].from);
	printf("</td>\n");
	printf("<td>");
	printf("%ld", message_size(i));
	printf("</td>\n");
	printf("<td>");
	printf("<a href=\"%s?mode=readmsg.html&msg=%d&ticket=", SCRIPT, i+1);
	printhtml(stdout, ticket);
	printf("\">\n");
	printheader(stdout, headers[i].subject);
	printf("</a>");
	printf("</td>\n");
	printf("</tr>\n");
}

void html_folders(char *f)
{
	char b[1000];
	DIR *dp;
	struct dirent *de;
	printf("<select name=newfolder>\n");
	csnprintf(b, sizeof b, "%s/%s/%s", SPOOLDIR, domain, username);
	dp = opendir(b);
	if (dp) {
		while ((de = readdir(dp))) {
			char *p;
			if ((de->d_name[0] != '_')
				&& !isalnum(de->d_name[0])) continue;
			if (strchr(de->d_name, '@')) continue;
			strncpy(b, de->d_name, sizeof b);
			p = strrchr(b, '.');
			if (p && !strcmp(p, ".BAK")) continue;
			if (p && !strcmp(p, ".gz")) *p = '\0';
			if (!strcmp(f, b)) {
				printf("<option value=\"%s\" selected>%s\n", b, b);
			} else {
				printf("<option value=\"%s\">%s\n", b, b);
			}
		}
		closedir(dp);
	}
	printf("</select><br>\n");
}

static void mailbox_filename(char *b, size_t bs, char *folder)
{
	blacklist(domain, "/@", '_');
	blacklist(username, "/.@", '_');
	downcase(username);
	csnprintf(b, bs, "%s/%s/%s", SPOOLDIR, domain, username);
	mkdir(b, 0700);
	csnprintf(b+strlen(b), bs-strlen(b), "/%s", folder);
}

/*
Read a mail file, store relevant header information.
nmsg is the number of messages in the file.
*/
static void parsemail(char *fn)
{
	char b[10000], more[1000];
	int c;
	int state = HEADER;
	FILE *fp = zopen(fn, "r");

	nmsg = 0;
	headers[nmsg].from = NULL;
	headers[nmsg].to = NULL;
	headers[nmsg].cc = NULL;
	headers[nmsg].subject = NULL;
	headers[nmsg].date = NULL;
	headers[nmsg].replyto = NULL;
	headers[nmsg].msgid = MSGID;
	headers[nmsg].ctype = NULL;
	headers[nmsg].ctenc = NULL;
	headers[nmsg].offset = ftell(fp);
	headers[0].offset = 0;
	if (fp == NULL) {
		html_error(_("Can't open mail file"));
	}
	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		if (!strcmp(b, ".")) {
			state = HEADER;
			nmsg++;
			headers[nmsg].from = NULL;
			headers[nmsg].to = NULL;
			headers[nmsg].subject = NULL;
			headers[nmsg].date = NULL;
			headers[nmsg].replyto = NULL;
			headers[nmsg].msgid = MSGID;
			headers[nmsg].ctype = NULL;
			headers[nmsg].ctenc = NULL;
			headers[nmsg].offset = ftell(fp);
		}
		switch (state) {
		case HEADER:
			/* icky stuff for continued header lines */
			c = getc(fp);
			ungetc(c, fp);
			if ((c == ' ' || c == '\t') &&
			    (sizeof b) > strlen(b)+(sizeof more)) {
				fgets(more, sizeof more, fp);
				chomp(more);
				strcat(b, more);
				c = getc(fp);
				ungetc(c, fp);
			}
			if (b[0] == '\0') {
				state = BODY;
			} else if (!cstrncasecmp(b, "Subject: ", 9)) {
				headers[nmsg].subject = dupstring(b+9);
			} else if (!cstrncasecmp(b, "From: ", 6)) {
				headers[nmsg].from = dupstring(b+6);
			} else if (!cstrncasecmp(b, "To: ", 4)) {
				if (headers[nmsg].to)
					headers[nmsg].to =
						catstrings3(headers[nmsg].to,
							", ", b+4);
				else
					headers[nmsg].to =
						dupstring(b+4);
			} else if (!cstrncasecmp(b, "Cc: ", 4)) {
				if (headers[nmsg].cc)
					headers[nmsg].cc =
						catstrings3(headers[nmsg].cc,
							", ", b+4);
				else
					headers[nmsg].cc =
						dupstring(b+4);
			} else if (!cstrncasecmp(b, "Date: ", 6)) {
				headers[nmsg].date = dupstring(b+6);
			} else if (!cstrncasecmp(b, "Reply-To: ", 10)) {
				headers[nmsg].replyto = dupstring(b+10);
			} else if (!cstrncasecmp(b, "Message-ID: ", 12)) {
				headers[nmsg].msgid = dupstring(b+12);
			} else if (!cstrncasecmp(b, "Content-Type: ", 14)) {
				headers[nmsg].ctype = dupstring(b+14);
			} else if (!cstrncasecmp(b, "Content-Transfer-Encoding: ", 27)) {
				headers[nmsg].ctenc = dupstring(b+27);
			}
			break;
		case BODY:
			break;
		}
	}
	fclose(fp);
}

/* Return message with this id, or 0 if not found. */
static int find_msgid(char *msgid)
{
	int i;

	for (i = 0; i < nmsg; i++)
		if (!strcmp(msgid, headers[i].msgid)) return i+1;
	return 0;
}

/* List all message headers in the mail file. Must run parsemail first. */
static void html_list(char *p, char *q)
{
	int i;
	char fn[1000];
	mailbox_filename(fn, sizeof fn, folder);
	parsemail(fn);
	printf(_("<h3>%d messages in %s.</h3>"), nmsg, folder);
	printf("\n<form method=post action=\"%s\">\n", SCRIPT);
	printf("<input type=hidden name=ticket value=%s>\n", ticket);
	printf("<table cellpadding=0 cellspacing=0 width=\"100%%\">\n");
	printf("<tr bgcolor=\"#333366\">\n");
	printf("<td align=\"center\" nowrap><font color=\"#ffffff\">"
		"&nbsp;</font></td>\n");
	printf("<td align=\"center\" nowrap><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Date"));
	printf("<td align=\"center\" nowrap><font color=\"#ffffff\">"
		"%s</font></td>\n", _("From"));
	printf("<td align=\"center\" nowrap><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Size"));
	printf("<td align=\"center\" nowrap><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Subject"));
	printf("</tr>\n");
	for (i = 1; i <= nmsg; i++) {
		html_top(i);
	}
	printf("</table><hr size=1 noshade>\n");
	html_folders(folder);
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Move marked"));
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Delete marked"));
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Delete folder"));
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Go to folder"));
	printf("</form>\n");

	printf( "<a href=\"%s?ticket=%s&mode=compose.html\">", SCRIPT, ticket);
	icon("new.gif", _("New message"));
	printf("</a>\n");

	printf("<a href=\"%s?ticket=%s&mode=fldmgr.html\">", SCRIPT, ticket);
	icon("fld_open.gif", _("New folder"));
	printf("</a><br>\n");

	if (adminprivs) {
#if 0
		printf("<a href=\"%s?mode=editsites.html&ticket=%s\">",
			SCRIPT, ticket);
		icon("nntp.gif", _("Edit sites"));
		printf("</a>\n");

		printf("<a href=\"%s?mode=editcx.html&ticket=%s\">",
			SCRIPT, ticket);
		icon("nntp.gif", _("Edit contexts"));
		printf("</a>\n");
#else
		printf("<a href=\"%s?mode=editdomain.html&ticket=%s\">",
			SCRIPT, ticket);
		icon("nntp.gif", _("Edit domain"));
		printf("</a>\n");
#endif
	}
	if (superprivs) {
		printf("<a href=\"%s?mode=newdomain.html&ticket=%s\">",
			SCRIPT, ticket);
		icon("nntp.gif", _("Add domain"));
		printf("</a>\n");
	}
	if (adminprivs || superprivs) printf("<br>\n");
}

int compar(const void *p, const void *q)
{
	const int *d1 = p;
	const int *d2 = q;

	return *d1 - *d2;
}

/*
Remove message n from the mail file. Del is an array of message numbers
to delete, ndel is the number of numbers.

Again, this function tries to parse the mailbox itself and fails.
Instead we will do it this way:

 - Sort the del array
 - Loop over the size of the mailbox, copying to fn or fn2 depending on offset

Much more efficient, and less prone to error.
*/
static void deletemail(char *fn, char *fn2, int *del, int ndel)
{
	FILE *fpin, *fpout1, *fpout2;
	char tmpname[1000];
	int deleted = 0;
	int o, d, od, c, size;
	if (fn2) {
		fpout2 = zopen(fn2, "a");
		if (fpout2 == NULL) {
			printf("%s\n", _("Can't open destination folder."));	
			return;
		}
	} else {
		fpout2 = NULL;
	}
	csnprintf(tmpname, sizeof tmpname, "%s.BAK", fn);
	rename(fn, tmpname);
	fpin = zopen(tmpname, "r");
	if (fpin == NULL) {
		printf("%s\n", _("Can't open source folder."));
		fclose(fpout2);
		rename(tmpname, fn);
		return;
	}
	fpout1 = zopen(fn, "w");
	if (fpout1 == NULL) {
		printf("%s\n", _("Can't open mail file."));
		fclose(fpin);
		fclose(fpout2);
		rename(tmpname, fn);
		return;
	}

	qsort(del, ndel, sizeof *del, compar);
	if ((ndel > 1) && (del[0] > del[1]))
		html_error("Broken sort!\n");

	o = 0;
	d = 0;

	/* portably get the size of the file */
#if 1
	size = headers[nmsg].offset;
#else
	fseek(fpin, 0, SEEK_END);
	size = ftell(fpin);
	rewind(fpin);
#endif

	while (d < ndel) {
		od = headers[del[d]-1].offset;	/* 1st byte of msg to del */
		/* keep everything up to do */
		while (o < od) {
			c = getc(fpin);
			putc(c, fpout1);
			o++;
		}
		od = headers[del[d]].offset; /* 1st byte after msg to move */
		/* move everything up to do */
		while (o < od) {
			c = getc(fpin);
			if (fpout2) putc(c, fpout2);
			o++;
		}
		d++;
	}
	/* pick up slack */
	while (o < size) {
		c = getc(fpin);
		putc(c, fpout1);
		o++;
	}

	deleted = ndel;

	fclose(fpin);
	fclose(fpout1);
	if (fpout2) {
		printf(_("Moved %d messages<p>%c"), deleted, '\n');
		fclose(fpout2);
	} else {
		printf(_("Deleted %d messages<p>%c"), deleted, '\n');
	}
}


/*
If quote is non-null, we add it at the beginning of lines.
If pretags is non-null, print headers and <pre> tags.

991229: use metamail to extract text and attachments rather than my own code.
*/
static void readmail(char *fn, int n, char *quote, char *pretags)
{
	FILE *fpin, *fp, *fpout;
	char *fnout, cmd[1024], attdir[1024];
	char e1[1024], e2[1024], e3[1024], e4[1024];
	int c, i;
	DIR *dp;
	struct dirent *de;

	csnprintf(e1, sizeof e1, "SHOHEIDIR=%s", SHOHEIDIR);
	putenv(e1);
	if (quote) {
		csnprintf(e2, sizeof e2, "QUOTE=%s", quote);
		putenv(e2);
	}
	if (pretags) {
		csnprintf(e4, sizeof e4, "PRETAGS=%s", pretags);
		putenv(e4);
	}
	csnprintf(attdir, sizeof attdir, "%s/%s", MIMEDIR, ticket);
	mkdir(attdir, 0700);
	csnprintf(e3, sizeof e3, "METAMAIL_TMPDIR=%s", attdir);
	putenv(e3);

	if (debugcgi) printf("readmail(%s, %d, %s)\n", fn, n, quote);

	if (n < 1 || n > nmsg) {
		char b[1000];
		csnprintf(b, sizeof b,
			_("Message number %d: no such message"), n);
		html_error(b);
	}
	fpin = zopen(fn, "r");
	if (fpin == NULL) {
		html_error(_("Can't open mail file."));
		return;
	}
	fseek(fpin, headers[n-1].offset, SEEK_SET);
	if (pretags) {
		printf("<table>\n");
		printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
			_("Date"), make_html(headers[n-1].date));
		printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
			_("From"), make_html(headers[n-1].from));
		printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
			_("To"), make_html(headers[n-1].to));
		if (headers[n-1].cc) {
			printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
				_("Cc"), make_html(headers[n-1].cc));
		}
		printf("<tr><td><b>%s:</b></td><td>", _("Subject"));
		printheader(stdout, headers[n-1].subject);
		printf("</td></tr>\n");
		printf("</table>\n");
		printf("<table width=\"100%%\" border=0 cellspacing=2 cellpadding=5>");
		printf("<tr>");
		printf("<td bgcolor=\"#eeeeee\">\n");
	}
	fnout = tmpnam(NULL);
	fpout = fopen(fnout, "w");
	if (fpout == NULL)
		html_error("Can't extract message");

	/* save whole message except trailing ".\n" */
	for (i = headers[n-1].offset; i < headers[n].offset-2; i++) {
		c = getc(fpin);
		putc(c, fpout);
	}
	fclose(fpout);
	csnprintf(cmd, sizeof cmd, "%s/readmime %s", SHOHEIDIR, fnout);
	fp = popen(cmd, "r");
	if (fp == NULL)
		html_error("Can't read message");
	while ((c = getc(fp)) != EOF)
		putchar(c);
	pclose(fp);
	fclose(fpin);
	if (!quote) {
		/* attachments */
		dp = opendir(attdir);
		if (dp) {
			while ((de = readdir(dp))) {
				if (de->d_name[0] == '.') continue;
				printf("%s: ", _("Attachment"));
				printf("<a href=\"%s/%s/%s\">%s</a><p>\n",
					MIMEURL, ticket,
					de->d_name, de->d_name);
			}
			closedir(dp);
		}
		printf("</td>");
		printf("</tr>");
		printf("</table>\n");
	}
}

static char *get_query(char *attr)
{
	int i;
	for (i = 0; i < nquery; i++)
		if (!strcmp(attr, query[i].attr)) return query[i].value;
	return NULL;
}

static char *require_query(char *attr)
{
	char *p = get_query(attr);
	if (p == NULL) {
		char b[1000];
		csnprintf(b, sizeof b,
			_("Missing required attribute %s"), attr);
		html_error(b);
	}
	return p;
}

/* translate the %NN stuff used in URLs into earthling characters */
static void decode_url(char *p)
{
	char *q = p;
	int c;

	while ((c = *p++)) {
		if (c == '+') {
			*q++ = ' ';
		} else if (c == '%') {
			sscanf(p, "%2x", &c);
			*q++ = c;
			if (*p) p++;
			if (*p) p++;
		} else {
			*q++ = c;
		}
	}
	*q = '\0';
}

/* opposite of above */
static char *encode_url(char *p, char *b, int bs)
{
	int c, i = 0;

	while ((c = *p++) && (i < bs-5)) {
		if (isalnum(c)) {
			b[i++] = c;
		} else {
			csnprintf(b+i, bs-i, "%%%02x", (c & 255));
			i += 3;
		}
	}
	b[i] = '\0';
	return b;
}

/*
Remove trailing cr and lf */
void rchomp(char *p)
{
	int n = strlen(p)-1;
	if (n && p[n] == '\n') p[n--] = '\0';
	if (n && p[n] == '\r') p[n--] = '\0';
}

#define CD "Content-Disposition: "

char *headsep(char *h)
{
	int state = 0;
	while (*h) {
		switch (state) {
		case 0:		/* looking for key */
			if (*h == ';') {
				*h = '\0';
				return h+1;
			}
			else if (*h == '"') state = 1;
			break;
		case 1:		/* in quotes */
			if (*h == '"') state = 0;
		}
		h++;
	}
	return NULL;	/* not found */
}

char *trimquotes(char *p)
{
	if (*p != '"') return p;
	p[strlen(p)-1] = '\0';
	return p+1;
}

/* look for name and filename in a content-disposition header */
void parseheader(char *h, char **n, char **fn)
{
	char *p;
	*n = *fn = NULL;
	while (h) {
		while (isspace(*h)) h++;
		p = headsep(h);
		if (!cstrncasecmp(h, "name=", 5))
			*n = trimquotes(h+5);
		else if (!cstrncasecmp(h, "filename=", 9))
			*fn = trimquotes(h+9);
		h = p;
	}
}

/*
This is for uploading a file.
*/
static void mimecgi(void)
{
	int i, n, msl;
	char *b, *p, *ms, *h, *name, *filename, *value;
	int cdl = strlen(CD);
	n = atoi(getenv("CONTENT_LENGTH"));
	b = cmalloc(n+1);
	fread(b, 1, n, stdin);
	i = 0;
	ms = b;
	p = memchr(b, '\n', n);
	if (!p) html_error(_("Error in form data"));
	*p = '\0';
	i += strlen(ms);	/* past mime separator */
	i++;			/* past the nul */
	chomp(ms);		/* remove cr */
	msl = strlen(ms);
	while (i < n) {
		/* read headers */
		h = b+i;
		p = memchr(h, '\n', n-i);
		if (!p) html_error(_("Error in form data"));
		*p = '\0';
		i += strlen(h);
		i++;
		chomp(h);
		while (*h) {
			if (!cstrncasecmp(h, CD, cdl)) {
				parseheader(h+cdl, &name, &filename);
			}
			h = b+i;
			p = memchr(h, '\n', n-i);
			if (!p) html_error(_("Error in form data"));
			*p = '\0';
			i += strlen(h);
			i++;
			chomp(h);
		}
		/* now we are at the data */
		value = b+i;
		while (strncmp(b+i, ms, msl)) i++;
		b[i] = '\0';
		rchomp(value);
		if (filename) {
			query[nquery].attr = name;
			query[nquery++].value = filename;
			filedata = value;
			filesize = i-2-(value-b);	/* 2 is for crlf */
		} else if (name) {
			query[nquery].attr = name;
			query[nquery++].value = value;
		}
		/* skip past mime separator */
		i += msl;
		/* check for crlf or -- */
		if (b[i] == '-') break;
		if (b[i] == '\r') i++;
		if (b[i] == '\n') i++;
	}
}

/*
Take all the stuff entered in forms and such and store it away in query[]
*/
static void parsecgi(void)
{
	char *method = getenv("REQUEST_METHOD");
	char *type = getenv("CONTENT_TYPE");
	char *p;
	int i;

	if (debugcgi) printf("parsecgi()\n");

	nquery = 0;
	if (method == NULL) html_template("login.html");
	if (!cstrcasecmp(method, "GET")) {
		char *qs = getenv("QUERY_STRING");
		if (qs == NULL) html_template("login.html");
		query_string = dupstring(qs);
	} else if (!cstrcasecmp(method, "POST")) {
		int n;
		char *ql = getenv("CONTENT_LENGTH");
		if (type && !cstrncasecmp(type, "multipart/", 10)) {
			mimecgi();
			return;
		}
		if (ql == NULL) html_template("login.html");
		n = atoi(ql);
		query_string = cmalloc(n+1);
		fread(query_string, 1, n, stdin);
		query_string[n] = '\0';
	} else {
		html_template("login.html");
	}
	for (p = strtok(query_string, "&"); p; p = strtok(NULL, "&")) {
		query[nquery++].attr = p;
	}
	for (i = 0; i < nquery; i++) {
		p = strchr(query[i].attr, '=');
		if (p) {
			*p = '\0';
			query[i].value = p+1;
			decode_url(query[i].value);
		} else {
			query[i].value = NULL;
		}
		if (debugcgi)
			printf("'%s'='%s'\n", query[i].attr, query[i].value);
	}
}

typedef struct {
	char *d;	/* domain name */
	char *n;	/* user-visible site name */
} dstruct;

static int dcompar(const void *p, const void *q)
{
	const dstruct *d1 = p;
	const dstruct *d2 = q;

	return strcmp(d1->n, d2->n);
}

/* More than 1000 sites would make the login screen really messy */
#define MAXSITES 1000

static void select_site(char *p, char *q)
{
	DIR *dp;
	FILE *fp;
	struct dirent *de;
	char b[1000];
	dstruct sites[MAXSITES];
	int nsites = 0;
	int i;
	printf("<select name=domain>\n");
	/* site list is generated dynamically and then sorted */
	dp = opendir(SPOOLDIR);
	if (dp == NULL)
		html_error(_("Can't open site list"));
	while ((nsites < MAXSITES) && ((de = readdir(dp)))) {
		if ((de->d_name[0] != '_')
			&& !isalnum(de->d_name[0])) continue;
		csnprintf(b, sizeof b, "%s/%s/@SITES", SPOOLDIR, de->d_name);
		fp = zopen(b, "r");
		if (fp) {
			while (fgets(b, sizeof b, fp)) {
				chomp(b);
				if (b[0]) {
					sites[nsites].d = dupstring(de->d_name);
					sites[nsites++].n = dupstring(b);
				}
			}
			fclose(fp);
		} else {
			sites[nsites].d = dupstring(de->d_name);
			sites[nsites++].n = dupstring(de->d_name);
		}
	}
	closedir(dp);
	qsort(sites, nsites, sizeof *sites, dcompar);
	for (i = 0; i < nsites; i++) {
		printf("<option value=\"%s\">%s\n",
			sites[i].d, sites[i].n);
	}

	printf(	"</select>");
}

static void select_lang(char *p, char *q)
{
	printf("<select name=language>\n"
		"<option value=sv>Svenska\n"
		"<option value=en>English\n"
		"</select>");
}

static void tmpl_variable(char *p, char *q)
{
	if (!strcmp(p, "SCRIPT")) printf("%s", SCRIPT);
	else if (!strcmp(p, "TICKET")) printf("%s", ticket);
	else html_error("Can't substitute variable");
}

static void tmpl_icon(char *p, char *q)
{
	char *r = strchr(q, '&');
	if (r) {
		*r++ = '\0';
		if (r[0] == '_') r = _(r+1);
	}
	icon(q, r);
}

static void tmpl_chkticket(char *p, char *q)
{
	chkticket();
}

static void tmpl_chkadmin(char *p, char *q)
{
	chkticket();
	if (!adminprivs) html_error(_("Not administrator"));
}

static void tmpl_chksuper(char *p, char *q)
{
	chkticket();
	if (!superprivs) html_error(_("Not supervisor"));
}

static void tmpl_header(char *p, char *q)
{
	if (!header_done) {
		header_done = 1;
		html_template("header.html");
	}
}

static void tmpl_footer(char *p, char *q)
{
	if (!footer_done) {
		footer_done = 1;
		html_template("footer.html");
	}
}

/* Different "modes" passed from the forms tell the program what to do */

static int try_help(char *cx)
{
	int c;
	FILE *fp;
	char *p, cx2[100], fn[2000];
	if (cx == NULL) cx = "main";
	strncpy(cx2, cx, sizeof cx2);
	if ((p = strstr(cx2, ".html"))) *p = '\0';
	blacklist(cx2, "/@", '_');
	if (!language) language = "en";
	csnprintf(fn, (sizeof fn)-200, "%s/%s.%s.html",
		HELPDIR, cx2, language);
	fp = zopen(fn, "r");
	if (!fp) return 0;
	while ((c = getc(fp)) != EOF) putchar(c);
	fclose(fp);
	return 1;
}

/*
Display context-sensitive help. Required attributes: none.
Optional attributes: cx.
*/
static void mode_help(char *p, char *q)
{
	char *cx = get_query("cx");
	ticket = get_query("ticket");
	if (!try_help(cx))
		try_help("main");
	printf("<p>\n");

/* Note: I would like to offer a "back" link on the help page, but it
is too difficult. It is not sufficient to link the way I do below,
because many required attributes such as msg will be missing. */
#if 0
	if (cx) {
		printf("<a href=\"%s?ticket=%s&mode=%s\">",
			SCRIPT, ticket, cx);
		icon("previous.gif", _("Back"));
		printf("</a>\n");
	}
#endif

	if (ticket) html_template("stdlinks.html");
}

/*
Display a message. Required attributes: ticket, msg.
*/
static void mode_read(char *p, char *q)
{
	char fn[1000];
	int no;
	msg = require_query("msg");
	no = atoi(msg);
	mailbox_filename(fn, sizeof fn, folder);
	parsemail(fn);
	if (no < 1 || no > nmsg)
		html_list(p, q);

	readmail(fn, no, NULL, "1");

	printf("<form method=post action=\"%s\">\n", SCRIPT);
	printf("<input type=hidden name=ticket value=%s>\n", ticket);
	printf("<input type=hidden name=msg value=%s>\n", msg);
	html_folders(folder);
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Move"));
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Delete folder"));
	printf("<input type=submit name=mode value=\"%s\">\n",
		_("Go to folder"));
	printf("</form>\n");

	printf("<a href=\"%s?ticket=%s&msg=%d&mode=delete.html\">",
		SCRIPT, ticket, no);
	icon("stop.gif", _("Delete"));
	printf("</a>\n");

	printf("<a href=\"%s?ticket=%s&msg=%d&mode=reply.html\">", SCRIPT, ticket, no);
	icon("redo.gif", _("Reply"));
	printf("</a>\n");

	printf("<a href=\"%s?ticket=%s&msg=%d&mode=replyall.html\">", SCRIPT, ticket, no);
	icon("redo.gif", _("Reply to all"));
	printf("</a>\n");

	printf("<a href=\"%s?ticket=%s&msg=%d&mode=forward.html\">", SCRIPT, ticket, no);
	icon("redo.gif", _("Forward"));
	printf("</a>\n");

	printf("<a href=\"%s?ticket=%s&mode=readmsg.html&msg=%d\">", SCRIPT, ticket, 1);
	icon("first.gif", _("First"));
	printf("</a>\n");

	if (no > 1) {
		printf("<a href=\"%s?ticket=%s&mode=readmsg.html&msg=%d\">",
			SCRIPT, ticket, no-1);
		icon("previous.gif", _("Previous"));
		printf("</a>\n");
	}

	if (no < nmsg) {
		printf("<a href=\"%s?ticket=%s&mode=readmsg.html&msg=%d\">",
			SCRIPT, ticket, no+1);
		icon("next.gif", _("Next"));
		printf("</a>\n");
	}

	printf("<a href=\"%s?ticket=%s&mode=readmsg.html&msg=%d\">",
		SCRIPT, ticket, nmsg);
	icon("last.gif", _("Last"));
	printf("</a>\n");
	printf("<br>\n");
}

/*
Extract something good to include as name in the phrase
"NAME wrote:"
Strategy:
1. Skip leading white space.
2. If the first character is alphanumeric, start there.
3. If the first character is a quote, start at the next character.
4. If the string contains a <, start at the next character.
5. Otherwise, return the string "You".
6. Copy everything up to but not including the first garbage character.
7. Strip trailing white space.
8. Return the result.
*/
static char *fromname(char *p)
{
	char *s, *t;

	for (s = p; *s && isspace(*s); s++);
	if (isalnum(*s));
	else if (*s == '"') s++;
	else if ((t = strchr(s, '<'))) s = t+1;
	else return dupstring("You");

	t = dupstring(s);
	for (s = t;
	     *s && (isalnum(*s) || strchr(".-_ ", *s) || (*s & 128));
	     s++);
	while (s > t && isspace(*(s-1))) s--;
	*s = '\0';
	return t;
}

/*
Reply to a message. Required attributes: ticket, msg.
If invoked as REPLY, reply to Reply-to or From. If invoked as REPLYALL,
reply to CC as well. If invoked as FORWARD, leave To: and CC:
headers blank.
*/
static void mode_reply(char *p, char *q)
{
	char fn[1000], *into, *cc, *oldsubject, *re, insubject[1000];
	char *quote = QUOTE;
	int no;
	msg = require_query("msg");
	no = atoi(msg);
	mailbox_filename(fn, sizeof fn, folder);
	parsemail(fn);
	if (headers[no-1].subject) oldsubject = headers[no-1].subject;
	else oldsubject = _("your mail");
	if (!cstrncasecmp(oldsubject, RE, strlen(RE))) re = "";
	else re = RE;
	if (headers[no-1].replyto)
		into = make_html(headers[no-1].replyto);
	else if (headers[no-1].from)
		into = make_html(headers[no-1].from);
	else
		into = "nobody";
	if (!cstrcasecmp(p, "REPLYALL") && headers[no-1].cc)
		cc = make_html(headers[no-1].cc);
	else
		cc = "";
	if (!cstrcasecmp(p, "FORWARD")) {
		into = cc = "";
		re = "Fwd: ";
		quote = "";
	}
	csnprintf(insubject, sizeof insubject,
		"%s%s", re, make_html(oldsubject));
	printf(	"<table>\n"
	       "<tr><td><b>%s:</b></td>"
		"<td><input type=text size=65 name=to value=\"%s\"></td></tr>\n",
		_("To"), into);
	printf( "<tr><td><b>%s:</b></td>"
		"<td><input type=text size=65 name=subject value=\"",
		_("Subject"));
	printheader(stdout, insubject);
	printf("\"></td></tr>\n");
	printf( "<tr><td>%s</td>"
		"<td><input type=text size=65 name=cc value=\"%s\"></td></tr>\n"
		"<tr><td>%s</td>"
		"<td><input type=text size=65 name=bcc></td></tr>\n"
		"</table>\n",
		_("Cc:"), cc, _("Bcc:"));
	printf("<input type=hidden name=mode value=send.html>\n"
	       "<input type=hidden name=ticket value=%s>\n"
	       "<hr size=1 noshade>\n"
	       "<textarea name=message cols=64 rows=10 wrap=\"physical\">\n",
		ticket);
	if (cstrcasecmp(p, "FORWARD")) {
		printf(_("%s wrote:"), fromname(headers[no-1].from));
		printf("\n\n");
	}
	readmail(fn, no, quote, NULL);
	printf("</textarea>\n");
}

/*
Reply to a message. Required attributes: ticket, msg, group.
*/
static void nntp_reply(char *p, char *q)
{
	char *group;
	char *oldsubject = "No subject", insubject[1000], *re;
	int no;
	msg = require_query("msg");
	no = atoi(msg);
	group = require_query("group");
	if (!cstrncasecmp(oldsubject, RE, strlen(RE))) re = "";
	else re = RE;
	csnprintf(insubject, sizeof insubject,
		"%s%s", re, make_html(oldsubject));
	printf("<table>\n");
	printf("<tr><td><b>%s:</b></td>"
	       "<td><input type=text size=65 name=group value=\"%s\"></td></tr>\n",
	       _("Newsgroups"), group);
	printf("<tr><td><b>%s:</b></td>"
	       "<td><input type=text size=65 name=subject value=\"",
	       _("Subject"));
	printheader(stdout, insubject);
	printf("\"></td></tr>\n");
	printf("</table>\n");
	printf("<input type=hidden name=mode value=nntpsend.html>\n"
	       "<input type=hidden name=ticket value=%s>\n"
	       "<hr size=1 noshade>\n"
	       "<textarea name=message cols=64 rows=10 wrap=\"physical\">\n",
	       ticket);
	printf("%s\n", _("You wrote:"));
	printf("%s%s\n", QUOTE, "Can't quote yet");
	printf("</textarea>\n");
}

/*
Required attributes: ticket, group, message, subject.
*/
static void nntp_send(char *p, char *q)
{
	char *group, *message, *subject;
	FILE *fp;
	group = require_query("group");
	message = require_query("message");
	subject = require_query("subject");
	fp = popen(INEWS, "w");
	if (fp == NULL) html_error(_("Can't post message"));
	fprintf(fp, "From: <%s@%s>\n", username, domain);
	fprintf(fp, "Newsgroups: %s\n", group);
	fprintf(fp, "Subject: %s\n", subject);
	fprintf(fp, "\n");
	fputs(message, fp);
	pclose(fp);
}

static int sevenbitclean(char *p)
{
	int c;

	while ((c = *p++)) {
		if (c & 0x80) return 0;
	}
	return 1;
}

/*
Send the completed message. Required attributes: ticket, to, message.
Optional attributes: cc, bcc, subject, attachment.
This can come from the forms produced by mode_compose or mode_reply above.
Afterwards we display the list of messages in INBOX.
*/
static void mode_send(char *p, char *q)
{
	char *to, *cc, *bcc, *subject, *message, *a, *attachment;
	FILE *fp, *mfp;
	char cmd[1024], b[1000], intext[1024], inattach[1024], outmail[1024];
	int c, i, pid = getpid();

	to = require_query("to");
	message = require_query("message");
	cc = get_query("cc");
	bcc = get_query("bcc");
	subject = get_query("subject");

	attachment = get_query("attachment");
	if (!attachment || !attachment[0] || !filedata || filesize == 0)
		attachment = NULL;
		
	fp = popen(SENDMAIL, "w");
	if (fp == NULL) html_error(_("Can't send message"));
	fprintf(fp, "From: <%s@%s>\n", username, domain);
	for (a = strtok(dupstring(to), ","); a; a = strtok(NULL, ","))
		fprintf(fp, "To: %s\n", a);
	if (cc && cc[0]) {
		for (a = strtok(dupstring(cc), ","); a; a = strtok(NULL, ","))
			fprintf(fp, "Cc: %s\n", a);
	}
	if (bcc && bcc[0]) {
		for (a = strtok(dupstring(bcc), ","); a; a = strtok(NULL, ","))
			fprintf(fp, "Bcc: %s\n", a);
	}
	if (subject && subject[0]) fprintf(fp, "Subject: %s\n", subject);

	if (attachment) {
		csnprintf(intext, sizeof intext, "/tmp/intext.%d", pid);
		csnprintf(inattach, sizeof inattach, "/tmp/inattach.%d", pid);
		csnprintf(outmail, sizeof outmail, "/tmp/outmail.%d", pid);

		mfp = fopen(intext, "w");
		if (!mfp) html_error(_("Can't send mail"));
		for (i = 0; (c = message[i]); i++)
			if (c != '\r') fputc(c, mfp);
		if (c != '\n') fputc('\n', mfp);
		fclose(mfp);

		mfp = fopen(inattach, "w");
		if (!mfp) html_error(_("Can't save attachment"));
		fwrite(filedata, 1, filesize, mfp);
		fclose(mfp);
		/* remove back- (DOS) or forward (Unix) slashes */
		if ((a = strrchr(attachment, '\\'))) attachment = a+1;
		if ((a = strrchr(attachment, '/'))) attachment = a+1;
		csnprintf(cmd, sizeof cmd,
			"metasend -b -f %s -m text/plain -n "
				"-f %s -m '%s; name=\"%s\"' "
				"-o %s",
			intext, inattach,
			mimename(b, sizeof b, attachment, "application/unknown"),
			attachment, outmail);
		system(cmd);

		mfp = fopen(outmail, "r");
		if (!mfp) html_error(_("Can't open mail"));
		while (fgets(b, sizeof b, mfp))
			fputs(b, fp);
		fclose(mfp);
#if 1
		remove(intext);
		remove(inattach);
		remove(outmail);
#endif
	} else if (sevenbitclean(message)) {
		fprintf(fp, "\n");
		fputs(message, fp);
	} else {
		csnprintf(intext, sizeof intext, "/tmp/intext.%d", pid);
		csnprintf(outmail, sizeof outmail, "/tmp/outmail.%d", pid);

		mfp = fopen(intext, "w");
		if (!mfp) html_error(_("Can't send mail"));
		for (i = 0; (c = message[i]); i++)
			if (c != '\r') fputc(c, mfp);
		if (c != '\n') fputc('\n', mfp);
		fclose(mfp);

		csnprintf(cmd, sizeof cmd,
			"metasend -b -f %s -m text/plain "
				"-o %s",
			intext, outmail);
		system(cmd);

		mfp = fopen(outmail, "r");
		if (!mfp) html_error(_("Can't open mail"));
		while (fgets(b, sizeof b, mfp))
			fputs(b, fp);
		fclose(mfp);
#if 1
		remove(intext);
		remove(outmail);
#endif
	}
	pclose(fp);
}

/*
Delete a message. Required attributes: ticket, msg.
Optional attribute: newfolder (defaults to deleted-messages).
Afterwards list the contents of the mailbox.
*/
static void mode_movemsg(char *p, char *q)
{
	char fn[1000], fn2[1000], *fld;
	int no;
	msg = require_query("msg");
	fld = get_query("newfolder");
	if (!fld) fld = "deleted-messages";
	no = atoi(msg);

	mailbox_filename(fn, sizeof fn, folder);
	mailbox_filename(fn2, sizeof fn2, fld);
	parsemail(fn);
	deletemail(fn, fn2, &no, 1);
	printf(_("Deleted message %d.<p>"), no);
}

/*
Delete a message. Required attributes: ticket, msg.
Afterwards list the contents of the mailbox.
*/
static void mode_delete(char *p, char *q)
{
	char fn[1000], fn2[1000];
	int no;
	msg = require_query("msg");
	no = atoi(msg);
#if 0
	msgid = require_query("msgid");
	if (cstrcasecmp(msgid, headers[no-1].msgid))
		html_error("Message ID does not match");
#endif

	mailbox_filename(fn, sizeof fn, folder);
	mailbox_filename(fn2, sizeof fn2, TRASH);
	parsemail(fn);
	deletemail(fn, fn2, &no, 1);
	printf(_("Deleted message %d.<p>"), no);
}

/* Return the named header from a message, or NULL if not found */
static char *pop_header(int msg, char *h)
{
	int n = strlen(h), done = 0;
	char b[1024], *p = NULL;

	csnprintf(b, sizeof b, "TOP %d 0\r\n", msg);
	write_data(b);
	read_data(b);
	if (b[0] != '+') {
		html_error(_("Couldn't get header"));
	}
	for (;;) {
		read_data(b);
		if (!strcmp(b, ".")) {
			break;		/* end of message */
		} else if (!b[0]) {
			done = 1;	/* end of headers */
		} else if (!done && !cstrncasecmp(b, h, n)) {
			p = dupstring(b+n);
		}
	}
	return p;
}

/* Get a message from the pop server and save it in a file */
static int pop_get(int msg, FILE *fp)
{
	char b[1024];
	csnprintf(b, sizeof b, "RETR %d\r\n", msg);
	write_data(b);
	read_data(b);
	if (b[0] != '+') {
		html_error(_("Couldn't get message"));
		return 0;
	}
	for (;;) {
		read_data(b);
		fprintf(fp, "%s\n", b);
		if (!strcmp(b, ".")) break;
	}
	return 1;
}

/* Delete a message. */
static int pop_delete(int msg)
{
	char b[1024];
	csnprintf(b, sizeof b, "DELE %d\r\n", msg);
	write_data(b);
	read_data(b);
	if (b[0] != '+') return 0;
	return 1;
}

/* Print a list of available messages on stdout. Return number of messages. */
static int pop_list(int verbose)
{
	char b[1024];
	int n;

	write_data("LIST\r\n");
	read_data(b);
	if (b[0] != '+') {
		if (verbose) printf(_("No new messages.<p>"));
		return 0;
	}
	for (n = 0; n < 10000; n++) {
		int size;
		read_data(b);
		if (!strcmp(b, ".")) break;
		sscanf(b, "%d %d", &msgno[n], &size);
	}
	return n;
}

/* Try to log in. Return 1 for success, 0 for failure. */
static int pop_login(char *username, char *password, int verbose)
{
	char b[1024];

	csnprintf(b, sizeof b, "USER %s\r\n", username);
	write_data(b);
	read_data(b);
	if (b[0] != '+') {
		if (verbose) printf(_("Bogus username: %s"), username);
		return 0;
	}
	csnprintf(b, sizeof b, "PASS %s\r\n", password);
	write_data(b);
	read_data(b);
	if (b[0] != '+') {
		if (verbose) printf(_("Bogus password"));
		return 0;
	}
	return 1;
}

/* return 1 if success, 0 otherwise */
static int try_login(char *host, char *user, char *pass, int verbose)
{
	char b[1024], *msgid;
	int i, n, id;
	int newmsg;
	FILE *fp;

	open_socket(host, POP3_PORT);
	read_data(b);
	n = pop_login(user, pass, verbose);
	if (!n) {
		close_socket();
		return 0;
	}
	newmsg = pop_list(1);
	mailbox_filename(b, sizeof b, INBOX);

	/* make sure that INBOX exists */
	fp = zopen(b, "a");
	fclose(fp);
	parsemail(b);		/* to get message ids */

	fp = zopen(b, "a");	/* append */
	if (fp == NULL) html_error(_("Can't open mailbox for writing"));
	for (i = 0; i < newmsg; i++) {
		msgid = pop_header(msgno[i], "Message-ID: ");
		id = find_msgid(msgid);
		if (!id)
			pop_get(msgno[i], fp);
		if (!leaveonserver)
			pop_delete(msgno[i]);
	}
	fclose(fp);
	write_data("QUIT\r\n");
	read_data(b);
	close_socket();
	return 1;
}

static void tmpl_query(char *p, char *q)
{
	char *v = get_query(q);
	if (v) fputs(v, stdout);
}

static void tmpl_require(char *p, char *q)
{
	char *v = require_query(q);
	fputs(v, stdout);
}

/* list a group, get xover and make links to articles */
static void nntp_group(char *p, char *q)
{
	char b[2024];
	int i, nmsg, reply, first, last;
	char *group;
	char msgno[1000], subject[1000], from[1000], date[1000], id[1000],
		ref[1000], bytes[1000], lines[1000], xref[1000];

	open_socket(NNTPSERVER, NNTP_PORT);
	read_data(b);
	if (b[0] != '2') html_error("bogus reply from nntp server");
	group = require_query("group");
	csnprintf(b, sizeof b, "GROUP %s\r\n", group);
	write_data(b);
	read_data(b);
	if (b[0] != '2') {
		csnprintf(b, sizeof b, "Can't change group to '%s'", group);
		html_error(b);
	}
	sscanf(b, "%d %d %d %d", &reply, &nmsg, &first, &last);
#if 0	/* bogus */
	printf(_("%d messages in %s<p>"), nmsg, group);
	printf("\n");
#endif
	csnprintf(b, sizeof b, "XOVER %d-%d\r\n", first, last);
	write_data(b);
	read_data(b);
	if (b[0] != '2') {
		csnprintf(b, sizeof b, "Can't get article list for %s", q);
		html_error(b);
	}
	printf("<table cellpadding=0 cellspacing=0 width=\"100%%\">\n");
	printf("<tr bgcolor=\"#333366\">\n");
	printf("<td align=\"center\"><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Date"));
	printf("<td align=\"center\"><font color=\"#ffffff\">"
		"%s</font></td>\n", _("From"));
#if 0
	printf("<td align=\"center\"><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Size"));
#endif
	printf("<td align=\"center\"><font color=\"#ffffff\">"
		"%s</font></td>\n", _("Subject"));
	printf("</tr>\n");
	read_data(b);
	i = 0;
	while (b[0] != '.') {
		*msgno = *subject = *from = *date = *id
			= *ref = *bytes = *lines = *xref = '\0';
		sscanf(b, "%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[\t]\t%s",
			msgno, subject, from, date, id, ref, bytes, lines, xref);

		printf("<tr bgcolor=\"%s\">\n", (i&1)?"#eeeeee":"#dddddd");
		printf("<td nowrap>");
		printhtml(stdout, short_date(date));
		printf("</td>\n");
		printf("<td>");
		printhtml(stdout, from);
		printf("</td>\n");
#if 0
		printf("<td>");
		printhtml(stdout, bytes);
		printf("</td>\n");
#endif
		printf("<td>");
		printf("<a href=\"%s?mode=nntpread.html&ticket=%s&group=%s&msg=%s\">",
			SCRIPT, ticket, group, msgno);
		printhtml(stdout, subject);
		printf("</a></td>\n");
		printf("</tr>\n");
		read_data(b);
		i++;
	}
	printf("</table>\n");
	write_data("QUIT\r\n");
	read_data(b);
	close_socket();

	printf("<a href=\"%s?mode=nntpcompose.html&ticket=%s&group=%s\">",
		SCRIPT, ticket, group);
	icon("new.gif", _("New message"));
	printf("</a>\n");

	printf("<a href=\"%s?mode=help.html&ticket=%s\">", SCRIPT, ticket);
	icon("nntp.gif", _("Help"));
	printf("</a>\n");
}

static void nntp_read(char *p, char *q)
{
	char b[1024];
	char *group;
	char *msg;
	char msgno[1000], subject[1000], from[1000], date[1000], id[1000],
		ref[1000], bytes[1000], lines[1000], xref[1000];

	open_socket(NNTPSERVER, NNTP_PORT);
	read_data(b);
	if (b[0] != '2') html_error("bogus reply from nntp server");
	group = require_query("group");
	msg = require_query("msg");
	csnprintf(b, sizeof b, "GROUP %s\r\n", group);
	write_data(b);
	read_data(b);
	if (b[0] != '2') {
		csnprintf(b, sizeof b,
			_("Can't change group to '%s'"), group);
		html_error(b);
	}

	csnprintf(b, sizeof b, "XOVER %s\r\n", msg);
	write_data(b);
	read_data(b);
	if (b[0] != '2') html_error("Can't get xover info");
	read_data(b);
	chomp(b);
	*msgno = *subject = *from = *date = *id
		= *ref = *bytes = *lines = *xref = '\0';
	sscanf(b, "%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[^\t]\t%[\t]\t%s",
		msgno, subject, from, date, id, ref, bytes, lines, xref);
	printf("<table>\n");
	printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
		_("Date"), make_html(date));
	printf("<tr><td><b>%s:</b></td><td>%s</td></tr>\n",
		_("From"), make_html(from));
	printf("<tr><td><b>%s:</b></td><td>", _("Subject"));
	printheader(stdout, subject);
	printf("</td></tr>\n");
	printf("</table>\n");
	read_data(b);	/* the single dot */

	csnprintf(b, sizeof b, "BODY %s\r\n", msg);
	write_data(b);
	read_data(b);
	if (b[0] != '2') {
		csnprintf(b, sizeof b,
			_("Can't get article %s in group '%s'"), msg, group);
		html_error(b);
	}
	read_data(b);
	printf("<table width=\"100%%\" border=0 cellspacing=2 cellpadding=5>\n");
	printf("<tr><td bgcolor=\"#eeeeee\">\n");
	printf("<pre>\n");
	while (strcmp(b, ".")) {
		printhtml(stdout, b);
		putchar('\n');
		read_data(b);
	}
	printf("</pre>\n");
	printf("</td></tr></table>\n");
	write_data("QUIT\r\n");
	read_data(b);
	close_socket();

	printf("<a href=\"%s?mode=nntpreply.html&msg=%s&ticket=%s&group=%s\">",
		SCRIPT, msg, ticket, group);
	icon("reply.gif", _("Reply"));
	printf("</a>\n");

	printf("<a href=\"%s?mode=help.html&ticket=%s\">", SCRIPT, ticket);
	icon("info.gif", _("Help"));
	printf("</a>\n");
}

static void mode_checkmbox(char *p, char *q)
{
	folder = INBOX;
	if (!try_login(hostname, dn, password, 1))
		html_error(_("Authentication failure"));
	writeticket();
}

/*
The user has entered name and password in a form. We must now try
to authenticate and, if successful, save the name and password in
a ticket file. Required attributes: username, password, domain.
List INBOX when done.

Users tried:
1. If @DN exists, try the username stored there
2. If 1 fails and @CX exists, append all contexts to username
3. If 1 and 2 fail, use username as it is

In all cases, a succesful login leads to @DN being rewritten and an
unsuccessful login leads to it being deleted.
*/
static void mode_login(char *p, char *q)
{
	char fn[1000], cx[1000];
	char b[1024], n[1024];
	FILE *fp;
	int done = 0;

	if (get_query("leaveonserver")) leaveonserver = 1;
	else leaveonserver = 0;
	username = require_query("username");
	if (username[0] == '\0')
		html_error(_("Null username not allowed"));
	blacklist(username, "@/.", '_');
	password = require_query("password");
	if (password[0] == '\0')
		html_error(_("Empty password not allowed"));
	domain = require_query("domain");
	if (domain[0] == '\0')
		html_error(_("Empty domain not allowed"));
	blacklist(domain, "@/", '_');
	hostname = server_lookup(domain);
	if (hostname == NULL)
		html_error(_("Can't map site to host name"));

	/* When we start, username is a plain username like "ulric"
	   and dn is undefined.
	   When we finish, username will still be the same, but dn
	   will contain the distinguished name of the user. For
	   NetWare 4/5 users, this may look like .ulric.bdpc.edu.stockholm,
	   while users in a flat user database such as Unix or NW 3
	   will have dn same as username.

	   This is important! We need both the "plain" username
	   and the distinguished name. The plain name is used
	   for the user directory name and for identifying admins.
	   The dn is used for logging into the pop server.
	*/
	/* try @DN */
	csnprintf(fn, sizeof fn, "%s/%s/%s/@DN", SPOOLDIR, domain, username);
	fp = zopen(fn, "r");
	if (fp) {
		fgets(dn, sizeof dn, fp);
		chomp(dn);
		if (try_login(hostname, dn, password, 0)) {
#if 0
			username = dupstring(dn);
#endif
			done = 1;
		}
		fclose(fp);
	}

	/* try @CX */
	if (!done) {
		csnprintf(fn, sizeof fn, "%s/%s/@CX", SPOOLDIR, domain);
		fp = zopen(fn, "r");
		if (fp) {
			while (fgets(cx, sizeof cx, fp)) {
				chomp(cx);
				csnprintf(dn, sizeof dn,
					".%s.%s", username, cx);
				if (try_login(hostname, dn, password, 0)) {
#if 0
					username = dupstring(dn);
#endif
					done = 1;
					break;
				}
			}
			fclose(fp);
		}
	}

	/* try plain name */
	if (!done) {
		strncpy(dn, username, sizeof dn);
		if (!try_login(hostname, dn, password, 1))
			html_error(_("Authentication failure"));
	}

	/* At this point, we know both plain username and dn. Bingo. */

	/* store @DN for future use */
	csnprintf(fn, sizeof fn,
		"%s/%s/%s/@DN", SPOOLDIR, domain, username);
	fp = zopen(fn, "w");
	fputs(dn, fp);
	fclose(fp);

	folder = INBOX;
	language = require_query("language");
	if (language == NULL) language = LANGUAGE;
	if (get_query("useicons")) useicons = 1;
	else useicons = 0;
	
	/* check if this is an administrator */
	adminprivs = 0;
	csnprintf(fn, sizeof fn, "%s/%s/@ADMINS", SPOOLDIR, domain);
	fp = zopen(fn, "r");
	if (fp) {
		while (fgets(b, sizeof b, fp)) {
			chomp(b);
			if (!cstrcasecmp(b, username)) {
				adminprivs = 1;
				break;
			}
		}
		fclose(fp);
	}
	/* check if this is a supervisor */
	superprivs = 0;
	csnprintf(fn, sizeof fn, "%s/@SUPERVISORS", SPOOLDIR);
	fp = zopen(fn, "r");
	if (fp) {
		csnprintf(n, sizeof n, "%s@%s", username, domain);
		while (fgets(b, sizeof b, fp)) {
			chomp(b);
			if (!cstrcasecmp(b, n)) {
				superprivs = 1;
				break;
			}
		}
		fclose(fp);
	}
	writeticket();
}

/*
The user has asked us to logout by removing the file containing the ticket.
Required attribute: ticket.
*/
static void mode_logout(char *p, char *q)
{
	char b[1000];
	chkticket();
	csnprintf(b, sizeof b, "%s/%s", TICKETDIR, ticket);
	remove(b);
	mailbox_filename(b, sizeof b, TRASH);
	remove(b);
}

/*
Move checked messages into a different folder. Required attributes:
ticket, folder. Optional attributes: msgN (one for each message to
process.
*/
static void mode_mmove(char *p, char *q)
{
	char fn[1000], *fld, fn2[1000];
	int *del;
	int i, ndel;
	fld = require_query("newfolder");
	mailbox_filename(fn, sizeof fn, folder);
	mailbox_filename(fn2, sizeof fn2, fld);
	parsemail(fn);
	del = cmalloc(nmsg*sizeof(int));
	ndel = 0;
	for (i = nmsg; i >= 1; i--) {
		char a[100];
		csnprintf(a, sizeof a, "msg%d", i);
		if (get_query(a)) del[ndel++] = i;
	}
	deletemail(fn, fn2, del, ndel);
	free(del);
}

/*
Delete checked messages.  Required attribute: ticket.
Optional attributes: msgN (one for each message to process)
*/
static void mode_mdel(char *p, char *q)
{
	char fn[1000], fn2[1000];
	int *del;
	int i, ndel;
	mailbox_filename(fn, sizeof fn, folder);
	mailbox_filename(fn2, sizeof fn2, TRASH);
	parsemail(fn);
	del = cmalloc(nmsg*sizeof(int));
	ndel = 0;
	for (i = nmsg; i >= 1; i--) {
		char a[100];
		csnprintf(a, sizeof a, "msg%d", i);
		if (get_query(a)) del[ndel++] = i;
	}
	deletemail(fn, fn2, del, ndel);
	free(del);
}

/*
Create mail folder.  Required attribute: newfolder.
*/
static void mode_newfld(char *p, char *q)
{
	char fn[1000], *fld;
	FILE *fp;
	fld = require_query("newfolder");
	mailbox_filename(fn, sizeof fn, fld);
	fp = zopen(fn, "a");
	fclose(fp);
#if 0
	folder = dupstring(fld);
	writeticket();
#endif
}

static void tmpl_setfolder(char *p, char *q)
{
	folder = dupstring(q);
	writeticket();
}

/*
Print @CX file on stdout. Required attributes: ticket.
*/
static void tmpl_editcx(char *p, char *q)
{
	char fn[1024], b[1024];
	FILE *fp;
	csnprintf(fn, sizeof fn, "%s/%s/@CX", SPOOLDIR, domain);
	fp = zopen(fn, "r");
	if (fp) {
		while (fgets(b, sizeof b, fp)) {
			if (b[0] != '\n' && b[0] != '\r')
				fputs(b, stdout);
		}
		fclose(fp);
	} else {
		puts(domain);
	}
}

/*
Save @CX file. Required attributes: cx, ticket.
*/
static void tmpl_savecx(char *p, char *q)
{
	char fn[1024], *cx;
	FILE *fp;
	cx = require_query("cx");
	csnprintf(fn, sizeof fn, "%s/%s/@CX", SPOOLDIR, domain);
	fp = zopen(fn, "w");
	if (fp) {
		fputs(cx, fp);
		fclose(fp);
	} else {
		printf("Can't save %s\n", fn);
	}
}

/*
Print @SITES file on stdout. Required attribute: ticket.
*/
static void tmpl_editsites(char *p, char *q)
{
	char fn[1024], b[1024];
	FILE *fp;
	csnprintf(fn, sizeof fn, "%s/%s/@SITES", SPOOLDIR, domain);
	fp = zopen(fn, "r");
	if (fp) {
		while (fgets(b, sizeof b, fp))
			if (b[0] != '\n' && b[0] != '\r')
				fputs(b, stdout);
		fclose(fp);
	} else {
		puts(domain);
	}
}

/*
Save @SITES file. Required attributes: sites, ticket.
*/
static void tmpl_savesites(char *p, char *q)
{
	char fn[1024], *sites;
	FILE *fp;
	sites = require_query("sites");
	csnprintf(fn, sizeof fn, "%s/%s/@SITES", SPOOLDIR, domain);
	fp = zopen(fn, "w");
	if (fp) {
		fputs(sites, fp);
		fclose(fp);
	} else {
		printf("Can't save %s\n", fn);
	}
}

/*
Create new domain. Required attributes: name, server.
Optional attributes: admins, cx, sites.
*/
static void tmpl_makedomain(char *p, char *q)
{
	int n;
	char fn[1024];
	FILE *fp;
	char *name = require_query("name");
	char *server = require_query("server");
	char *admins = get_query("admins");
	char *cx = get_query("cx");
	char *sites = get_query("sites");
	csnprintf(fn, sizeof fn, "%s/%s", SPOOLDIR, name);
	n = mkdir(fn, 0700);
	if (n != 0) {
		if (errno == EEXIST)
			printf(_("Domain %s exists. "), name);
		html_error(_("Can't create domain"));
	}
	csnprintf(fn, sizeof fn, "%s/%s/@SERVER", SPOOLDIR, name);
	fp = zopen(fn, "w");
	if (fp == NULL) html_error(_("Can't save @SERVER"));
	fputs(server, fp);
	fclose(fp);
	if (admins && admins[0]) {
		csnprintf(fn, sizeof fn, "%s/%s/@ADMINS", SPOOLDIR, name);
		fp = zopen(fn, "w");
		if (fp == NULL) html_error(_("Can't save @ADMINS"));
		fputs(admins, fp);
		fclose(fp);
	}
	if (cx && cx[0]) {
		csnprintf(fn, sizeof fn, "%s/%s/@CX", SPOOLDIR, name);
		fp = zopen(fn, "w");
		if (fp == NULL) html_error(_("Can't save @CX"));
		fputs(cx, fp);
		fclose(fp);
	}
	if (sites && sites[0]) {
		csnprintf(fn, sizeof fn, "%s/%s/@SITES", SPOOLDIR, name);
		fp = zopen(fn, "w");
		if (fp == NULL) html_error(_("Can't save @SITES"));
		fputs(sites, fp);
		fclose(fp);
	}
}

static int daysofmonth(int year, int month)
{
	year += 1900;
	switch (month) {
	case 0: case 2: case 4: case 6: case 7: case 9: case 11:
		return 31;
	case 3: case 5: case 8: case 10:
		return 30;
	default:
		if ((year % 400) == 0) return 29;
		if ((year % 100) == 0) return 28;
		if ((year % 4) == 0) return 29;
		return 28;
	}
}

static char *monthnames[] = {
	"January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December"
};

static char *monthname(int m)
{
	if (m < 0 || m > 11) return "out of bounds";
	return monthnames[m];
}

/*
Display calendar for one month. Required attribute: ticket.
Optional attributes: year, month.
*/
static void mode_calendar(char *p, char *q)
{
	time_t ttime;
	struct tm *ltime, now;
	int d, md, wd, today;
	char *year = get_query("year");
	char *month = get_query("month");
	char fn[1024];
	FILE *fp;

	ttime = time(NULL);
	ltime = localtime(&ttime);
	now = *ltime;

	if (year) {
		ltime->tm_year = atoi(year);
		ltime->tm_mday = 1;
	}
	if (month) {
		ltime->tm_mon = atoi(month);
		ltime->tm_mday = 1;
	}

	printf("<a href=\"%s?mode=calendar.html&ticket=%s&",
		SCRIPT, ticket);
	if (ltime->tm_mon == 0)
		printf("year=%d&month=11\">", ltime->tm_year-1);
	else
		printf("year=%d&month=%d\">", ltime->tm_year, ltime->tm_mon-1);
	printf("&lt;&lt;</a>\n");
	printf("<b>%s %d</b>", _(monthname(ltime->tm_mon)), 1900+ltime->tm_year);
	printf("<a href=\"%s?mode=calendar.html&ticket=%s&",
		SCRIPT, ticket);
	if (ltime->tm_mon == 11)
		printf("year=%d&month=0\">", ltime->tm_year+1);
	else
		printf("year=%d&month=%d\">", ltime->tm_year, ltime->tm_mon+1);
	printf("&gt;&gt;</a><p>\n");

	ltime->tm_mday = 1;
	ttime = mktime(ltime);

	printf("<table>\n");
	printf("<tr>\n");
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Su"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Mo"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Tu"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("We"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Th"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Fr"));
	printf("<td bgcolor=\"#333366\"><font color=\"#ffffff\">%s</td>\n", _("Sa"));
	printf("</tr>\n");

	if (ltime->tm_wday != 0) {
		printf("<tr>\n");
		for (wd = 0; wd < ltime->tm_wday; wd++) {
			printf("<td bgcolor=\"#cccccc\">&nbsp;</td>\n");
		}
	} else {
		wd = 0;
	}
	md = daysofmonth(ltime->tm_year, ltime->tm_mon);
	d = 1;
	while (d <= md) {
		today = ((ltime->tm_year == now.tm_year)
			&& (ltime->tm_mon == now.tm_mon)
			&& (d == now.tm_mday));
		if (wd == 0) printf("<tr>\n");
		printf("<td bgcolor=\"#cccccc\">");
		printf("<a href=\"%s?mode=editday.html&ticket=%s&year=%d&month=%d&day=%d\">",
			SCRIPT, ticket, ltime->tm_year, ltime->tm_mon, d);
		csnprintf(fn, sizeof fn, "%s/%s/%s/@CAL/%d/%d/%d",
			SPOOLDIR, domain, username,
			ltime->tm_year, ltime->tm_mon, d);
		fp = zopen(fn, "r");
		if (fp) printf("<i>");
		if (today) printf("<b>");
		printf("%d", d);
		if (today) printf("</b>");
		if (fp) {
			printf("</i>");
			fclose(fp);
		}
		printf("</a>");
		printf("</td>\n");
		d++;
		wd = (wd+1) % 7;
		if (wd == 0) printf("</tr>\n");
	}
	if (wd) {
		while (wd < 7) {
			printf("<td bgcolor=\"#cccccc\">&nbsp;</td>\n");
			wd++;
		}
		printf("</tr>\n");
	}
	printf("</table>\n");

	printf("<p>\n");
}

/*
Edit appointments for one day. Required attributes: ticket, year, month, day.
*/
static void mode_editday(char *p, char *q)
{
	char fn[1024];
	char *y = require_query("year");
	char *m = require_query("month");
	char *d = require_query("day");
	int year, month, day;
	FILE *fp;
	int c;

	year = atoi(y);
	if (year < 0 || year > 3000)
		html_error("bogus year");
	month = atoi(m);
	if (month < 0 || month > 11)
		html_error("bogus month");
	day = atoi(d);
	if (day < 1 || day > daysofmonth(year, month))
		html_error("bogus month");

	printf("<h2>%d %s</h2>\n", day, _(monthname(month)));
	printf("<form method=post action=\"%s\">\n", SCRIPT);
	printf("<textarea name=app cols=60 rows=10 wrap=\"physical\">\n");
	csnprintf(fn, sizeof fn, "%s/%s/%s/@CAL/%d/%d/%d",
		SPOOLDIR, domain, username, year, month, day);
	fp = zopen(fn, "r");
	if (fp) {
		while ((c = getc(fp)) != EOF) putchar(c);
		fclose(fp);
	}
	printf("</textarea>\n");
	printf("<br>\n");
	printf("<input type=hidden name=mode value=saveday.html>\n");
	printf("<input type=hidden name=year value=%d>\n", year);
	printf("<input type=hidden name=month value=%d>\n", month);
	printf("<input type=hidden name=day value=%d>\n", day);
	printf("<input type=hidden name=ticket value=%s>\n", ticket);
	printf("<input type=submit value=OK>\n");
	printf("<br>\n");
	printf("<a href=\"%s?ticket=%s&mode=clrday.html&year=%d&month=%d&day=%d\">",
		SCRIPT, ticket, year, month, day);
	printf("%s\n", _("Delete"));
	printf("</form>\n");
}

/*
Save appointments for one day. Required attributes: ticket, year,
month, day, app.
*/
static void mode_saveday(char *p, char *q)
{
	char fn[1024];
	char *y = require_query("year");
	char *m = require_query("month");
	char *d = require_query("day");
	char *app = require_query("app");
	int year, month, day;
	FILE *fp;

	year = atoi(y);
	if (year < 0 || year > 3000)
		html_error("bogus year");
	month = atoi(m);
	if (month < 0 || month > 11)
		html_error("bogus month");
	day = atoi(d);
	if (day < 1 || day > daysofmonth(year, month))
		html_error("bogus month");

	csnprintf(fn, sizeof fn, "%s/%s/%s/@CAL", SPOOLDIR, domain, username);
	mkdir(fn, 0700);
	csnprintf(fn+strlen(fn), (sizeof fn)-strlen(fn), "/%d", year);
	mkdir(fn, 0700);
	csnprintf(fn+strlen(fn), (sizeof fn)-strlen(fn), "/%d", month);
	mkdir(fn, 0700);
	csnprintf(fn+strlen(fn), (sizeof fn)-strlen(fn), "/%d", day);
	fp = zopen(fn, "w");
	if (fp == NULL)
		html_error("Can't save day\n");

	fputs(app, fp);
	fclose(fp);
}

/*
Clear appointments for one day. Required attributes: ticket, year,
month, day.
*/
static void mode_clrday(char *p, char *q)
{
	char fn[1024];
	char *y = require_query("year");
	char *m = require_query("month");
	char *d = require_query("day");
	int year, month, day;

	year = atoi(y);
	if (year < 0 || year > 3000)
		html_error("bogus year");
	month = atoi(m);
	if (month < 0 || month > 11)
		html_error("bogus month");
	day = atoi(d);
	if (day < 1 || day > daysofmonth(year, month))
		html_error("bogus month");

	csnprintf(fn, sizeof fn, "%s/%s/%s/@CAL/%d/%d/%d",
		SPOOLDIR, domain, username, year, month, day);
	remove(fn);
}

/*
List all notes for one user. Required attributes: ticket.
*/
static void mode_listnotes(char *p, char *q)
{
	char fn[1024], b[1024], b2[1024], b3[1024], *p1;
	FILE *fp;
	DIR *dp;
	struct dirent *de;

	csnprintf(fn, sizeof fn,
		"%s/%s/%s/@NOTES", SPOOLDIR, domain, username);
	mkdir(fn, 0700);
	dp = opendir(fn);
	if (dp) {
		while ((de = readdir(dp))) {
			if ((de->d_name[0] != '_')
				&& !isalnum(de->d_name[0])) continue;
			strncpy(b, de->d_name, sizeof b);
			p1 = strrchr(b, '.');
			if (p1 && !strcmp(p1, ".BAK")) continue;
			if (p1 && !strcmp(p1, ".gz")) *p1 = '\0';
			csnprintf(b3, sizeof b3,
				"<a href=\"%s?mode=editnote.html&notename=%s"
				"&ticket=%s\">%s</a>",
				SCRIPT, encode_url(b, b2, sizeof b2),
				ticket, b);
fp = fopen("/tmp/grr", "w");
fputs(b3, fp);
fclose(fp);
			fputs(b3, stdout);
			printf("<br>\n");
		}
		closedir(dp);
	}
	printf("<p>");
	printf("<a href=\"%s?ticket=%s&mode=newnote.html\">", SCRIPT, ticket);
	icon("new.gif", _("New note"));
	printf("</a><br>\n");
}

/*
Edit a note. Required attributes: ticket, notename.
*/
static void mode_editnote(char *p, char *q)
{
	char fn[1024], b2[1024];
	char *notename = require_query("notename");
	FILE *fp;
	int c;

	blacklist(notename, "@/", '_');
	printf("<form method=post action=\"%s\">\n", SCRIPT);
	printf("<textarea name=note cols=60 rows=10 wrap=\"physical\">\n");
	csnprintf(fn, sizeof fn, "%s/%s/%s/@NOTES/%s",
		SPOOLDIR, domain, username, notename);
	fp = zopen(fn, "r");
	if (fp) {
		while ((c = fgetc(fp)) != EOF) putchar(c);
		fclose(fp);
	}
	printf("</textarea>\n");
	printf("<br>\n");
	printf("<input type=hidden name=mode value=\"savenote.html\">\n");
	printf("<input type=hidden name=ticket value=\"%s\">\n", ticket);
	printf("<input type=hidden name=notename value=\"%s\">\n", notename);
	printf("<input type=submit value=OK>\n");
	printf("</form>\n");
	printf("<a href=\"%s?mode=delnote.html&ticket=%s&\n", SCRIPT, ticket);
	printf("notename=%s", encode_url(notename, b2, sizeof b2));
	printf("\">%s</a><br>\n", _("Delete note"));
}

/*
Save a note. Required attributes: ticket, notename, note.
*/
static void mode_savenote(char *p, char *q)
{
	char fn[2000];
	char *notename = require_query("notename");
	char *note = require_query("note");
	FILE *fp;

	blacklist(notename, "@/", '_');
	csnprintf(fn, (sizeof fn)-100,
		"%s/%s/%s/@NOTES", SPOOLDIR, domain, username);
	mkdir(fn, 0700);
	strcat(fn, "/");
	strncat(fn, notename, 100);
	fp = zopen(fn, "w");
	if (fp == NULL) html_error("Can't save note");

	fputs(note, fp);
	fclose(fp);
}

/*
Delete a note. Required attributes: ticket, notename.
*/
static void mode_delnote(char *p, char *q)
{
	char fn[1024];
	char *notename = require_query("notename");

	blacklist(notename, "@/", '_');
	csnprintf(fn, sizeof fn, "%s/%s/%s/@NOTES/%s",
		SPOOLDIR, domain, username, notename);
	remove(fn);
}

/*
Create a folder. Required attributes: newfolder.
*/
static void mode_delfld(char *p, char *q)
{
	char fn[1000], *fld, dfn[1000];
	FILE *fp, *dfp;
	int c;
	fld = require_query("newfolder");
	mailbox_filename(fn, sizeof fn, fld);
	mailbox_filename(dfn, sizeof dfn, TRASH);
	if (strcmp(fld, TRASH)) {
		if ((fp = zopen(fn, "r"))) {
			if ((dfp = zopen(dfn, "a"))) {
				while ((c = getc(fp)) != EOF)
					putc(c, dfp);
				fclose(dfp);
			}
			fclose(fp);
		}
	}
	remove(fn);
}

/*
Go to folder. Required attributes: newfolder.
*/
static void mode_gotofld(char *p, char *q)
{
	folder = require_query("newfolder");
	writeticket();
}

/*
Delete a web file. Required attributes: filename.
*/
static void www_delete(char *p, char *q)
{
	char *a;
	char *filename = require_query("filename");
	char fn[1024];
	if ((a = strrchr(filename, '\\'))) filename = a+1;
	if ((a = strrchr(filename, '/'))) filename = a+1;
	blacklist(filename, "@/", '_');
	csnprintf(fn, sizeof fn, "%s/%s/%s/@WWW/%s",
		SPOOLDIR, domain, username, filename);
	remove(fn);
}

/*
List the contents of the web page. Required attributes: none.
*/
static void www_list(char *p, char *q)
{
	char b[1024], b2[1024];
	DIR *dp;
	struct dirent *de;
	struct stat statbuf;
	csnprintf(b, sizeof b, "%s/%s/%s/@WWW", SPOOLDIR, domain, username);
	mkdir(b, 0700);
	printf("<table>\n");
	dp = opendir(b);
	if (dp) {
		printf("<table>\n");
		while ((de = readdir(dp))) {
			sprintf(b, "%s/%s/%s/@WWW/%s",
				SPOOLDIR, domain, username, de->d_name);
			if (stat(b, &statbuf) || !(statbuf.st_mode & S_IFREG))
				continue;
			printf("<tr>\n");
			if ((de->d_name[0] != '_')
				&& !isalnum(de->d_name[0])) continue;
			printf("<td>%s</td>\n", de->d_name);
			printf("<td><a href=\"%s?ticket=%s&mode=wwwdelete.html&filename=%s",
				SCRIPT, ticket, encode_url(de->d_name, b2, sizeof b2));
			printf("\">%s</a></td>\n", _("Delete"));
			printf("</tr>\n");
		}
		closedir(dp);
		printf("</table>\n");
	}
}

/*
Save a file in the home page. Required attribute: filename.
*/
static void www_save(char *p, char *q)
{
	FILE *fp;
	char *a, b[1024];
	char *filename = require_query("filename");
	if (!filename || !filedata || !filesize) {
		printf("Can't save file\n");
	} else {
		if ((a = strrchr(filename, '\\'))) filename = a+1;
		if ((a = strrchr(filename, '/'))) filename = a+1;
		blacklist(filename, "/@", '_');
		csnprintf(b, sizeof b, "%s/%s/%s/@WWW/%s",
			SPOOLDIR, domain, username, filename);
		fp = fopen(b, "w");
		if (!fp) {
			printf("Can't save file\n");
		} else {
			fwrite(filedata, 1, filesize, fp);
			fclose(fp);
		}
	}
}

static struct {
	char *name;
	void (*cmd)(char *, char *);
} substers[] = {
	{"HEADER", tmpl_header},
	{"FOOTER", tmpl_footer},
	{"SELECT_SITE", select_site},
	{"SELECT_LANG", select_lang},
	{"SCRIPT", tmpl_variable},
	{"TICKET", tmpl_variable},
	{"ICON", tmpl_icon},
	{"CHKTICKET", tmpl_chkticket},
	{"CHKADMIN", tmpl_chkadmin},
	{"CHKSUPER", tmpl_chksuper},
	{"CHECKMBOX", mode_checkmbox},
	{"MOVEMSG", mode_movemsg},
	{"DELETE", mode_delete},
	{"DELFLD", mode_delfld},
	{"FORWARD", mode_reply},
	{"GOTOFLD", mode_gotofld},
	{"HELP", mode_help},
	{"LISTFLD", html_list},
	{"LOGIN", mode_login},
	{"LOGOUT", mode_logout},
	{"MDEL", mode_mdel},
	{"MMOVE", mode_mmove},
	{"NEWFLD", mode_newfld},
	{"NNTPGROUP", nntp_group},
	{"NNTPREAD", nntp_read},
	{"NNTPREPLY", nntp_reply},
	{"NNTPSEND", nntp_send},
	{"QUERY", tmpl_query},
	{"READMSG", mode_read},
	{"REPLY", mode_reply},
	{"REPLYALL", mode_reply},
	{"REQUIRE", tmpl_require},
	{"SEND", mode_send},
	{"SETFOLDER", tmpl_setfolder},
	{"EDITCX", tmpl_editcx},
	{"SAVECX", tmpl_savecx},
	{"EDITSITES", tmpl_editsites},
	{"SAVESITES", tmpl_savesites},
	{"MAKEDOMAIN", tmpl_makedomain},
	{"CALENDAR", mode_calendar},
	{"EDITDAY", mode_editday},
	{"SAVEDAY", mode_saveday},
	{"CLRDAY", mode_clrday},
	{"LISTNOTES", mode_listnotes},
	{"EDITNOTE", mode_editnote},
	{"SAVENOTE", mode_savenote},
	{"DELNOTE", mode_delnote},
	{"WWWDELETE", www_delete},
	{"WWWLIST", www_list},
	{"WWWSAVE", www_save},
	{NULL, NULL},
};

static void html_subst(char *s)
{
	int i;
	char *p = strchr(s, '?');
	if (s[0] == '_') {
		fputs(_(s+1), stdout);
		return;
	} else if (s[0] == '\0') {
		putc('@', stdout);
	}
	if (p) *p++ = '\0';
	for (i = 0; substers[i].name; i++) {
		if (!strcmp(s, substers[i].name)) break;
	}
	if (substers[i].name) (*substers[i].cmd)(s, p);
	else html_template(s);
}

static void html_template(char *p)
{
	char fn[1000], s[1000];
	int c, state, i;
	FILE *fp;
	csnprintf(fn, sizeof fn, "%s/%s.%s", TEMPLATEDIR, p, language);
	fp = zopen(fn, "r");
	if (fp == NULL) {
		csnprintf(fn, sizeof fn, "%s/%s", TEMPLATEDIR, p);
		fp = zopen(fn, "r");
	}
	if (fp == NULL) {
		csnprintf(s, sizeof s, "Can't open HTML template %s", fn);
		html_error(s);
	}
	state = 0;
	i = 0;
	while ((c = getc(fp)) != EOF) {
		switch (state) {
		case 0:
			if (c == '@') {
				state = 1;
				i = 0;
			} else {
				putchar(c);
			}
			break;
		case 1:
			if (c == '@') {
				s[i] = '\0';
				state = 0;
				html_subst(s);
			} else {
				s[i++] = c;
			}
		}
	}
}

#ifdef HOMEPAGES
#define USER_PREFIX "/~"

/*
The argument fn is a filename. If fn is a regular file, send it. If fn
is a directory, try index.html in that directory followed by index.htm.

Returns 1 if a file was found, and 0 if not.
*/
static int serve_file(char *fn)
{
	char b[1000], mn[1000];
	char *m;
	int c, n;
	struct stat statbuf;
	FILE *fp;
	FILE *fpd = fopen("/tmp/homepages", "a");
	fprintf(fpd, "fn='%s'\n", fn);
	strncpy(b, fn, sizeof b);
	n = stat(b, &statbuf);
	if (n != 0) return 0;
	if (statbuf.st_mode & S_IFDIR) {
		csnprintf(b, sizeof b, "%s/index.html", fn);
		n = stat(b, &statbuf);
		if (n != 0) {
			csnprintf(b, sizeof b, "%s/index.htm", fn);
			n = stat(b, &statbuf);
		}
	}
	m = mimename(mn, sizeof mn, b, "text/plain");
	fprintf(fpd, "m='%s' b='%s'\n", m, b);
	fclose(fpd);
	fp = fopen(b, "r");
	if ((n != 0) || (statbuf.st_mode & S_IFREG) == 0 || fp == NULL) {
		return 0;
	}
	printf("Content-type: %s\r\n\r\n", m);
	while ((c = getc(fp)) != EOF)
		putchar(c);
	fclose(fp);
	return 1;
}

/*
A homepage URL looks something like:
	http://server.com/~username/filename.html
This will make REQUEST_URI look like /~username/filename.html
The file to return is SPOOLDIR/domain/username/@WWW/filename.html
This requires a virtual host in httpd.conf with DOCUMENT_ROOT
set up as SPOOLDIR/domain.
If there is no filename, or the file is a directory, try appending
index.html. So the URL http://server.com/~username
would try to return $DOCUMENT_ROOT/username/@WWW/index.html

r is DOCUMENT_ROOT, p is username/filename
*/
static void homepages(char *r, char *p)
{
	char fn[1000], b[1000];
	char *u, *f;
	FILE *fpd;
	if (r == NULL || p == NULL) return;
	fpd = fopen("/tmp/homepages", "a");
	fprintf(fpd, "%ld: REQUEST_URI='%s'\n", (long)time(NULL), p);
	strncpy(b, p, sizeof b);
	u = b;
	f = strchr(b, '/');
	if (f) {
		*f++ = '\0';
	} else {
		f = "index.html";
	}
	fprintf(fpd, "user='%s', file='%s'\n", u, f);
	fclose(fpd);
	csnprintf(fn, sizeof fn, "%s/%s/@WWW/%s", r, u, f);
	if (serve_file(fn)) exit(0);
}

/*
Similar to above, but handles request such as http://server.com/file.html.
Returns $DOCUMENT_ROOT/@WWW/file.html.
*/
static void sitepages(void)
{
	char fn[1000], b[1000];
	char *f;
	FILE *fpd;
	char *p = getenv("REQUEST_URI");
	char *r = getenv("DOCUMENT_ROOT");
	char *m = getenv("REQUEST_METHOD");
	if (p == NULL || p[0] != '/' || p[1] == '?') return;
	if (!m || !cstrcasecmp(m, "POST")) return;
	if (!strncmp(p, USER_PREFIX, strlen(USER_PREFIX))) {
		homepages(r, p+strlen(USER_PREFIX));
		return;
	}
	if (r == NULL) return;
	fpd = fopen("/tmp/homepages", "a");
	fprintf(fpd, "%ld: REQUEST_URI='%s'\n", (long)time(NULL), p);
	strncpy(b, p+1, sizeof b);
	f = b;
	fprintf(fpd, "file='%s'\n", f);
	fclose(fpd);
	csnprintf(fn, sizeof fn, "%s/@WWW/%s", r, f);
	if (serve_file(fn)) exit(0);
}
#endif

static void html_shohei_error(char *p)
{
   	html_error(p);
}

int main(void)
{
	char *mode;
	debugcgi = (getenv("DEBUGCGI") != NULL);
	if (debugcgi) printf("main()\n");

	umask(077);

   	shohei_error = html_shohei_error;
   
#ifdef HOMEPAGES
	sitepages();
#endif

	parsecgi();
	mode = get_query("mode");
	ticket = get_query("ticket");
	if (ticket) readticket();	/* this also sets language */
	else language = get_query("language");
	if (!language) language = LANGUAGE;
	i18n(language);

	if (mode == NULL) {
		html_template("login.html");
	} else if (!cstrcasecmp(mode, _("Move marked"))) {
		html_template("mmove.html");
	} else if (!cstrcasecmp(mode, _("Delete marked"))) {
		html_template("mdel.html");
	} else if (!cstrcasecmp(mode, _("Create folder"))) {
		html_template("newfld.html");
	} else if (!cstrcasecmp(mode, _("Delete folder"))) {
		html_template("delfld.html");
	} else if (!cstrcasecmp(mode, _("Go to folder"))) {
		html_template("gotofld.html");
	} else if (!cstrcasecmp(mode, _("Move"))) {
		html_template("movemsg.html");
	} else {
		html_template(mode);
	}
	return 0;
}

